import time
from dataclasses import dataclass
from enum import Enum
from typing import Any, Callable, List, Tuple, Union

from junit_xml import TestCase


class TestCaseCategories(Enum):
    FUNCTION = "function"
    FIXTURE = "fixture"


class MainRunner(Enum):
    PYTEST = "pytest"
    PYTHON = "python"
    NONE = "none"


class JunitCaseException(BaseException):
    def __init__(self, exception: BaseException, *args: object) -> None:
        super().__init__(*args)
        self.exception: BaseException = exception

    def __str__(self) -> str:
        return self.exception.__str__()

    def __repr__(self) -> str:
        return self.exception.__repr__()


@dataclass
class CaseFailure:
    message: str
    output: str = ""
    type: str = ""

    def __getitem__(self, item):
        return self.__getattribute__(item)


@dataclass
class TestCaseData:
    case: TestCase
    _func: Callable
    _start_time: float
    parametrize: Union[None, List[Tuple[str, Any]]] = None
    _has_parent = False
    is_inside_fixture = False
    had_exception = False

    @property
    def name(self):
        return self._func.__name__

    @property
    def has_parent(self):
        return self._has_parent

    def set_fin_time(self):
        self.case.elapsed_sec = time.time() - self._start_time

    def get_case_key(self):
        """ return immutable key """
        if self.parametrize:
            return tuple(sorted(self.parametrize))
        return ()

    def set_parametrize(self, params: List[Tuple[str, Any]]):
        self.parametrize = params
        self.case.name += f'[{", ".join([f"{p[0]}={p[1]}" for p in params])}]'

    def set_parent(self, case_parent_name: str):
        self._has_parent = True
        self.case.classname = f"{self.case.classname}.{case_parent_name}"



import re
from typing import Callable, List, Union

from _pytest.mark import Mark
from _pytest.python import Function

from ._junit_test_suite import JunitTestSuite


def is_suite_exist(obj: Union, func_name: str):
    if obj is None:
        return False

    func = getattr(obj, func_name)
    if hasattr(func, "__wrapped__"):
        return JunitTestSuite.is_suite_exist(func.__wrapped__)
    return False


def get_fixture_mark_function(own_markers: List[Mark], func: Function) -> Union[Callable, None]:
    """
    If mark parameterize decorate test suite with given fixture _get_mark_function is searching
    for all parametrize arguments and permutations.
    If mark type is parametrize, set self._parametrize with the current argument permutation.
    The representation of function name and arguments for N parametrize marks are as followed:
        func.name = some_func_name[param1-param2-...-paramN]
    :param own_markers: Pytest markers taken from func stack_locals
    :param func: wrapped function contained with pytest Function object
    :return: wrapped function itself
    """
    marks = [m for m in own_markers if m.name == "parametrize"]
    marks_count = len(marks)

    if marks_count == 0:
        return None

    params_regex = "-".join(["(.*?)"] * marks_count)
    args = list(re.compile(r"(.*?)\[{0}]".format(params_regex)).findall(func.name).pop())
    func_name = args.pop(0)

    params = list()
    for i in range(marks_count):
        params.append((marks[i].args[0], args[i]))

    return get_wrapped_function(func, func_name)


def get_wrapped_function(func: Function, func_name: str = None) -> Callable:
    """
    Get wrapped function from class or module
    :param func: wrapped function from class or module
    :param func_name: If there is pytest parameterized suite name is needed
    :return:
    """
    if func_name is None:
        func_name = func.name

    if func.cls:
        return getattr(func.cls, func_name).__wrapped__
    return getattr(func.module, func_name).__wrapped__



from contextlib import suppress
from types import GeneratorType
from typing import Callable, Union

import decorator

from ._junit_test_case import JunitTestCase, TestCaseCategories
from ._junit_test_suite import JunitTestSuite


class JunitFixtureTestCase(JunitTestCase):
    """
    This class defines TestCase for pytest fixtures.
    When use, need to make sure that the fixture come first, for example:
        @pytest.fixture
        @JunitFixtureTestCase()
        def my_fixture()
            ...
    """

    _generator: Union[GeneratorType, None]

    AFTER_YIELD_EXCEPTION_MESSAGE_PREFIX = "[TEARDOWN EXCEPTION]"

    def __init__(self) -> None:
        super().__init__()
        self._generator = None

    def __call__(self, function: Callable) -> Callable:
        self._func = function

        def wrapper(_, *args, **kwargs):
            value = self._wrapper(function, *args, **kwargs)
            yield value
            try:
                if self._generator:
                    self._teardown_yield_fixture(self._generator)
            except BaseException as e:
                self._add_failure(e, self.AFTER_YIELD_EXCEPTION_MESSAGE_PREFIX)
                raise
            finally:
                suite_key = self.get_suite_key()
                JunitTestSuite.fixture_cleanup(self._case_data, suite_key)

        return decorator.decorator(wrapper, function)

    @classmethod
    def _teardown_yield_fixture(cls, it) -> None:
        """Execute the teardown of a fixture function by advancing the iterator
        after the yield and ensure the iteration ends (if not it means there is
        more than one yield in the function)."""
        with suppress(StopIteration):
            next(it)

    def _execute_wrapped_function(self, *args, **kwargs):
        generator = super()._execute_wrapped_function(*args, **kwargs)
        try:
            self._generator = generator
            return next(generator)
        except (StopIteration, TypeError):
            return None

    def _on_wrapper_end(self):
        """
        Fixtures executing before the test suite, due to that issue the suite can't collect the fixture case
        if there is an exception during it's execution. If exception occur, it call JunitTestSuite class method
        collect_all that trigger the suite to collect all cases and export them into xml
        :return: None
        """
        self._case_data.case.category = TestCaseCategories.FIXTURE.value

        success = super(JunitFixtureTestCase, self)._on_wrapper_end()
        if not success:
            return success
        if len(self._case_data.case.failures) > 0:
            JunitTestSuite.collect_all()

        return success



import traceback
from contextlib import suppress
from typing import Any, Callable, Dict, List, Union

from _pytest.python import Function
from junit_xml import TestCase

from . import _utils
from ._junit_decorator import JunitDecorator
from ._junit_test_suite import JunitTestSuite
from ._test_case_data import CaseFailure, JunitCaseException, TestCaseCategories, TestCaseData


class JunitTestCase(JunitDecorator):
    """
    JunitTestCase is a decorator that represents single TestCase.
    When the decorated function finished its execution, it's registered to the relevant JunitTestSuite.
    TestCase will fail (TestCase failure) only when exception occurs during execution.
    """

    _func: Union[Callable, None]
    _stack_locals: List[Dict[str, Any]]
    _case_data: Union[TestCaseData, None]

    def __init__(self) -> None:
        super().__init__()
        self._case_data = None
        self._is_inside_fixture = False

    @property
    def name(self):
        if self._func:
            return self._case_data.name

    def _on_wrapper_start(self, function):
        super()._on_wrapper_start(function)
        case = TestCase(name=function.__name__, classname=self._get_class_name(),
                        category=TestCaseCategories.FUNCTION.value)
        self._case_data = TestCaseData(_start_time=self._start_time, case=case, _func=function)

    def _add_failure(self, e: BaseException, message_prefix: str = ""):
        message = f"{message_prefix} {str(e)}" if message_prefix else str(e)
        failure = CaseFailure(message=message, output=traceback.format_exc(), type=e.__class__.__name__)
        self._case_data.case.failures.append(failure)

    def _on_exception(self, e: BaseException):
        self._case_data.had_exception = True
        if isinstance(e, JunitCaseException):
            raise  # already registered on son test case
        self._add_failure(e)
        raise JunitCaseException(exception=e)

    def _on_wrapper_end(self) -> bool:
        self._case_data.set_fin_time()
        suite_func = self.get_suite_key()
        if suite_func is None:
            return False
        JunitTestSuite.register_case(self._case_data, suite_func)
        return True

    def get_suite_key(self) -> Union[Callable, None]:
        """
        Get suite function as unique key.
        This function handles also on case that the test suite function decorated with pytest.mark.parametrize and
        collect its parameters and record them into the test case.
        :return: Wrapped Suite function instance
        """
        with suppress(AttributeError):
            suite_func = self._get_suite()
            return suite_func

        return None

    def _set_parameterized(self, pytest_function: Function):
        parameterized = self.get_pytest_parameterized(pytest_function)

        if parameterized and not self._case_data.parametrize:
            self._case_data.set_parametrize(parameterized)

    def _get_suite(self):
        stack_locals = [stack_local for stack_local in self._stack_locals if "self" in stack_local]
        function_locals = [stack_local for stack_local in stack_locals if "function" in stack_local]

        self._set_parameterized(self._pytest_function)
        is_inside_fixture = False

        suite_func = self._get_function_suite(function_locals)
        if suite_func is None:
            is_inside_fixture = True
            suite_func = self._get_fixture_suite(self._pytest_function)
            # test case inside fixture with suite_func - if also exception - that is our case

        self._case_data.is_inside_fixture = is_inside_fixture
        self._is_inside_fixture = is_inside_fixture
        return suite_func

    def _get_function_suite(self, function_locals: List[Dict[str, Any]]):
        suite_func = None

        for f_locals in function_locals:
            suite = f_locals["self"]
            if isinstance(suite, JunitTestCase) and f_locals["function"].__name__ != self.name:
                if not self._case_data.has_parent:
                    self._case_data.set_parent(f_locals["function"].__name__)

            if isinstance(suite, JunitTestSuite):
                suite_func = f_locals["function"]
                break
        return suite_func

    @staticmethod
    def _get_fixture_suite(func: Function):
        with suppress(AttributeError):
            # if pytest.mark.parametrize exist, get actual function from class while ignoring the add parameters
            if hasattr(func, "own_markers") and len(func.own_markers) > 0:
                mark_function = _utils.get_fixture_mark_function(func.own_markers, func)
                if mark_function:
                    return mark_function
            return _utils.get_wrapped_function(func)

        return None
